"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.createLRU = void 0;
const createLRU = (options) => {
    let { max, onEviction } = options;
    if (!(Number.isInteger(max) && max > 0))
        throw new TypeError('`max` must be a positive integer');
    let size = 0;
    let head = 0;
    let tail = 0;
    let free = [];
    const keyMap = new Map();
    const keyList = new Array(max).fill(undefined);
    const valList = new Array(max).fill(undefined);
    const next = new Array(max).fill(0);
    const prev = new Array(max).fill(0);
    const moveToTail = (index) => {
        if (index === tail)
            return;
        const nextIndex = next[index];
        if (index === head)
            head = nextIndex;
        else {
            const prevIndex = prev[index];
            next[prevIndex] = nextIndex;
            prev[nextIndex] = prevIndex;
        }
        next[tail] = index;
        prev[index] = tail;
        tail = index;
    };
    const _evict = () => {
        const evictHead = head;
        const key = keyList[evictHead];
        onEviction === null || onEviction === void 0 ? void 0 : onEviction(valList[evictHead], key);
        keyMap.delete(key);
        keyList[evictHead] = undefined;
        valList[evictHead] = undefined;
        head = next[evictHead];
        size--;
        return evictHead;
    };
    return {
        /** Adds a key-value pair to the cache. Updates the value if the key already exists. */
        set(key, value) {
            let index = keyMap.get(key);
            if (index === undefined) {
                index = size === max ? _evict() : free.length > 0 ? free.pop() : size;
                keyMap.set(key, index);
                keyList[index] = key;
                size++;
            }
            else
                onEviction === null || onEviction === void 0 ? void 0 : onEviction(valList[index], key);
            valList[index] = value;
            moveToTail(index);
        },
        /** Retrieves the value for a given key and moves the key to the most recent position. */
        get(key) {
            const index = keyMap.get(key);
            if (index === undefined)
                return;
            moveToTail(index);
            return valList[index];
        },
        /** Retrieves the value for a given key without changing its position. */
        peek: (key) => {
            const index = keyMap.get(key);
            return index !== undefined ? valList[index] : undefined;
        },
        /** Checks if a key exists in the cache. */
        has: (key) => keyMap.has(key),
        /** Iterates over all keys in the cache, from least recent to most recent. */
        *keys() {
            let current = head;
            for (let i = 0; i < size; i++) {
                yield keyList[current];
                current = next[current];
            }
        },
        /** Iterates over all values in the cache, from least recent to most recent. */
        *values() {
            let current = head;
            for (let i = 0; i < size; i++) {
                yield valList[current];
                current = next[current];
            }
        },
        /** Iterates over `[key, value]` pairs in the cache, from least recent to most recent. */
        *entries() {
            let current = head;
            for (let i = 0; i < size; i++) {
                yield [keyList[current], valList[current]];
                current = next[current];
            }
        },
        /** Iterates over each key-value pair in the cache, from most recent to least recent. */
        forEach: (callback) => {
            let current = head;
            for (let i = 0; i < size; i++) {
                const key = keyList[current];
                const value = valList[current];
                callback(value, key);
                current = next[current];
            }
        },
        /** Deletes a key-value pair from the cache. */
        delete(key) {
            const index = keyMap.get(key);
            if (index !== undefined) {
                onEviction === null || onEviction === void 0 ? void 0 : onEviction(valList[index], key);
                keyMap.delete(key);
                free.push(index);
                keyList[index] = undefined;
                valList[index] = undefined;
                size--;
                return true;
            }
            return false;
        },
        /** Evicts the oldest item or the specified number of the oldest items from the cache. */
        evict: (number) => {
            let toPrune = Math.min(number, size);
            while (toPrune > 0) {
                _evict();
                toPrune--;
            }
        },
        /** Clears all key-value pairs from the cache. */
        clear() {
            for (const index of keyMap.values())
                onEviction === null || onEviction === void 0 ? void 0 : onEviction(valList[index], keyList[index]);
            keyMap.clear();
            keyList.fill(undefined);
            valList.fill(undefined);
            free = [];
            size = 0;
            head = tail = 0;
        },
        /** Resizes the cache to a new maximum size, evicting items if necessary. */
        resize: (newMax) => {
            if (!(Number.isInteger(newMax) && newMax > 0))
                throw new TypeError('`max` must be a positive integer');
            max = newMax;
            while (size > max)
                _evict();
        },
        /** Returns the maximum number of items that can be stored in the cache. */
        get max() {
            return max;
        },
        /** Returns the number of items currently stored in the cache. */
        get size() {
            return size;
        },
        /** Returns the number of currently available slots in the cache before reaching the maximum size. */
        get available() {
            return max - size;
        },
    };
};
exports.createLRU = createLRU;
